"""Interface with psql and capture all input and output.

Uses pexpect in combination with pyte for interfacing and screen-scraping
"""

from typing import Callable, List
from functools import reduce
import pexpect
from pyte import Stream, Screen
from shutil import get_terminal_size

from .psqlparser import PsqlParser


class PsqlWrapper:
    """Handles terminal interfacing with psql, using the parameter parser \
    to pick up relevant SQL statements for the hook function.

    Only one (static) class instance is intended for use in program.
    """

    debug: bool = False
    supported_psql_versions: list[str] = ["14.5"]

    def __init__(self, psql_args: bytes,
                 hook_f: Callable[[str], str], parser: PsqlParser):
        """Build wrapper for selected database.

        :param db_name_parameter: is name of database we are connecting to.
        :param hook_f: is a callback to which scraped SQL queries are passed \
        to and from which semantic error messages are received in return.
        :param parser: A parser that implements the required parsing functions.
        """
        self.psql_args: bytes = psql_args
        self.analyze: Callable[[str], str] = hook_f
        self.parser: PsqlParser = parser

        # shutil.get_terminal_size()
        (self.cols, self.rows) = get_terminal_size()

        # pyte.Screen, pyte.Stream
        self.pyte_screen: Screen = Screen(self.cols, self.rows)
        self.pyte_screen_output_sink: Stream = Stream(self.pyte_screen)

        # Analysis is always done when user presses Return
        # and resulting message is saved until when new prompt comes in
        self.pg4n_message: str = ""

    def start(self) -> None:
        """Start psql process and then start feeding hook function with \
        intercepted output."""

        version_msg = self.check_psql_version()
        if version_msg != "":
            print(version_msg)

        c = pexpect.spawn("psql " + bytes.decode(self.psql_args),
                          encoding="utf-8",
                          dimensions=(self.rows, self.cols))

        c.interact(input_filter=self.ifilter,
                   output_filter=self.ofilter)

    def check_psql_version(self) -> str:
        version_info = \
            pexpect.spawn("psql " + bytes.decode(self.psql_args) +
                          " --version")
        version_info.expect(pexpect.EOF)
        version_info_str: str = bytes.decode(version_info.before)
        version: str = self.parser.parse_psql_version(version_info_str)
        version_ok: bool = version in self.supported_psql_versions
        if version_ok:
            return ""
        else:
            return "Pg4n has only been tested on psql versions " + \
                str(self.supported_psql_versions) + "."
        

    def ifilter(self, input: bytes) -> bytes:
        """User input filter function for pexpect.interact: not used.

        :param input: user input characters.
        :returns: unchanged input.
        """
        return input

    def ofilter(self, output: bytes) -> bytes:
        """Forward output to `check_and_act_on_repl_output()` and feed \
        output to pyte screen for future screen-scraping.

        :param output: output seen on terminal screen.
        :returns: output with injected semantic error messages.
        """
        new_output: bytes = self.check_and_act_on_repl_output(output)

        self.pyte_screen_output_sink.feed(bytes.decode(new_output))

        if self.debug:
            f = open("pyte.screen", "w")
            f.write('\n'.join(
                line.rstrip() for line in self.pyte_screen.display))
            f.close()
            g = open("psqlwrapper.log", "a")
            g.write(str(new_output) + '\n')
            g.close()

        return new_output

    def check_and_act_on_repl_output(self, latest_output: bytes) -> bytes:
        """Check if user has hit Return so we can start analyzing, \
        or if a fresh prompt has come in and we can show them a helpful \
        message.

        :param latest_output: is used for Return press and fresh prompt \
        analysis. It is also what the helpful message will be injected to.
        :returns: output with injected semantic error messages.
        """
        # for optimization reasons, check output only if len() > 1, so keyboard
        # input does not trigger parsing (Return is always at least 2 length)
        if len(latest_output) <= 1:
            return latest_output

        # User hit Return: parse for potential SQL query, analyze, and
        # possibly provide a message to be included in next new prompt.
        if self._user_hit_return(latest_output):
            screen: str = '\n'.join(
                line.rstrip() for line in self.pyte_screen.display)
            parsed_sql_query: str = \
                self.parser.parse_last_found_stmt(screen)
            if parsed_sql_query != "":
                # feed query to hook function and save resulting message
                self.pg4n_message = self.analyze(parsed_sql_query)

        # If we have a semantic error message waiting and there is a fresh
        # prompt:
        # optimization: do not spend time parsing if there is no message:
        # AND-clause will stop executing after first False.
        if self.pg4n_message != "" \
           and self.parser.parse_new_prompt(
               bytes.decode(latest_output)) != []:  # there is new prompt
            new_output: bytes = self._replace_prompt(latest_output)
            self.pg4n_message = ""
            return new_output

        return latest_output

    def _replace_prompt(self, prompt: bytes) -> bytes:
        """Inject saved semantic error message into given prompt.

        :param prompt: is where the message is injected. A fresh prompt is \
        expected.
        :returns: a prompt with injected message.
        """
        split_prompt: List[str] = \
            self.parser.parse_new_prompt_and_rest(
                bytes.decode(prompt, "utf-8"))
        print_msg = self.pg4n_message.replace("\n", "\r\n")
        return bytes(split_prompt[0] + "\r\n" +
                     print_msg + "\r\n\r\n" +
                     split_prompt[1],
                     "utf-8")

    def _user_hit_return(self, output: bytes) -> bool:
        """Check if user hit return.

        :param output: is raw console output to be checked for return press.
        :returns: if user has indeed hit return.
        """
        # trivial case:
        # we assume only situation with \r\n at start of
        # line is when user has pressed enter
        if output[0:2] == b'\r\n':
            return True

        # complicated case: user has ctrl-R'd, copypasted command or something.
        # and the \r\n is somewhere in midst of output..
        if self.parser.parse_magical_return(
                bytes.decode(output, "utf-8")) != []:
            return True
        return False



"""Parse psql output."""

from typing import List
from pyparsing import \
    CaselessLiteral, Char, Combine, Literal, ParseException, \
    ParseResults, ParserElement, StringEnd, White, Word, ZeroOrMore, \
    identbodychars, nums
from functools import reduce
from string import printable

# Turn on memoization optimization
ParserElement.enablePackrat(None)

# default whitespace rules complicate things needlessly, remove them:
ParserElement.setDefaultWhitespaceChars('')


class PsqlParser:
    """Parses psql output for syntactic analysis."""

    debug: bool = False

    prompt_chars: str = identbodychars
    stmt_end: str = ";"
    # Naively, SQL statement body can contain all printable characters
    # (incl. whitespace), apart from ';'
    stmt_chars: str = \
        printable.translate(str.maketrans("", "", stmt_end))

    # rev means reversed, these are for performance reasons.
    # superusers have =# prompt

    tok_stmt_end: ParserElement = \
        Char(';')

    tok_rev_prompt_end: ParserElement = \
        Literal(" >=") | Literal(" #=")

    match_rev_any_sql_stmt: ParserElement = \
        ZeroOrMore(White()) + tok_stmt_end + ... + \
        tok_rev_prompt_end

    # TODO(?): We may have to match for errors
#    match_error: ParserElement = \
#        Literal("ERROR:")

    # magic strings related seemingly solely to ctrl-R use are
    # "\r\n\x1b[?2004l\r", "\r\n\r\r\n" and "\x08\r\n".
    match_rev_magical_returns: ParserElement = \
        Literal("\r\n\x1b[?2004l\r"[::-1]) \
        | Literal("\r\n\r\r\n"[::-1]) \
        | Literal("\x08\r\n"[::-1])

    def __init__(self):
        """Plain constructor for PsqlParser."""
        pass

    def parse_new_prompt(self, psql: str) -> List[str]:
        """Parse for an empty prompt, to detect when a query \
        evaluation has ended.

        :param psql: Raw console output that includes terminal control codes.
        :returns: an empty list if no prompt found. Otherwise has [' >='] or \
        [' #='].
        """
        # cheaper and easier to reverse & start from the end
        psql_rev: str = psql[::-1]
        results: List[str] = []
        prompt_res: ParseResults = None

        match_rev_prompt_end: ParserElement = \
            self.tok_rev_prompt_end

        try:
            prompt_res = match_rev_prompt_end.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if prompt_res:
            results = prompt_res.as_list()

        return results

    def parse_new_prompt_and_rest(self, psql: str) -> List[str]:
        """Parse for an empty prompt and everything preceding it \
        to 2 cells.

        :param psql: Raw console output that includes terminal control codes.
        :returns: a two-part list with everything before the prompt \
        line and then the prompt, to allow easy message injection.
        """
        # cheaper and easier to reverse & start from the end
        psql_rev: str = psql[::-1]
        results: List[str] = []
        prompt_res: ParseResults = None

        match_rev_prompt_and_then_rest: ParserElement = \
            self.tok_rev_prompt_end + \
            Word(self.prompt_chars) + \
            (StringEnd() | (Literal("?[\x1b") +
                            ... + StringEnd()))
        # ^either stops after dbname or includes \x1b[?2004l...

        try:
            prompt_res = match_rev_prompt_and_then_rest.parse_string(psql_rev)
        except ParseException as e:
            print(e.explain())
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if prompt_res:
            res_list = prompt_res.as_list()
            if len(res_list) == 4:  # results include \x1b[?2004l
                results = [res_list[3][::-1],
                           res_list[2][::-1] +
                           res_list[1][::-1] +
                           res_list[0][::-1]
                           ]
            elif len(res_list) == 2:  # parsing stops right after database name
                results = ['',
                           res_list[1][::-1] +
                           res_list[0][::-1]
                           ]

        return results

    def parse_magical_return(self, psql: str) -> List[str]:
        """Parse for weird Return presses.

        :param psql: Raw console output that includes terminal control codes.
        :returns: an empty list if no presses are found. \
        Otherwise has newline text.
        """
        psql_rev: str = psql[::-1]

        results: List[str] = []
        magical_return_res: ParseResults = None

        try:
            magical_return_res = \
                self.match_rev_magical_returns.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if magical_return_res:
            results = magical_return_res.as_list()
        return results

    def parse_last_found_stmt(self, psql: str) -> str:
        """Parse for last SQL query statement in a string.

        :param psql: screenscraped psql string

        :returns: parsed SQL query as plain string.
        """
        # reverse string for parsing efficiency
        psql_rev = psql[::-1]  # slicing is fastest operation for reversing

        # Match statement that might have \r\n or whitespace at the end.
        # Parse prompt text at the end, so multiline queries can be cleaned
        # properly.
        match_rev_last_stmt: ParserElement = \
            ZeroOrMore(Char("\n") | Char("\r") | White()) + \
            self.match_rev_any_sql_stmt + Word(self.prompt_chars)

        results: List[str] = []
        stmt_res: ParseResults = None
        db_name: str = ""

        try:
            stmt_res = match_rev_last_stmt.parse_string(psql_rev)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        # If parsing was successful, pick interesting parts.
        if stmt_res:
            stmt_res_list = stmt_res.as_list()
            length: int = len(stmt_res_list)

            results = [stmt_res_list[length - 3],
                       stmt_res_list[length - 4]]
            db_name = stmt_res_list[length - 1][::-1]

        # reverse back, concatenate, and remove \n's

        # TODO/BUG: removing \n's is a tough problem, see:
        # "pgdb=> SELECT * FROM orders WHERE order_tot
        # al_eur = 100;"
        # should parse as \n -> ""
        # "pgdb=> SELECT
        # * FROM orders WHERE order_total_eur = 100;
        # should parse as \n -> " " to avoid "SELECT* FROM ..".
        # replacing \n's with "" maybe has less edge cases. Or actually more.
        reversed_flattened_res: str = \
            reduce(lambda x, y: x + y[::-1], results, "")
        no_newlines_res = reversed_flattened_res.replace('\n', ' ')

        # Is the statement a SELECT statement?
        match_select_stmt: ParserElement = \
            ZeroOrMore(White()) + CaselessLiteral("SELECT")
        is_select: bool = False
        try:
            is_select = \
                match_select_stmt.parse_string(no_newlines_res) is not []
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        # If it is SELECT, remove multiline delimiters and then statement is
        # ready for analysis.
        if is_select:
            demultilined_res = no_newlines_res\
                .replace(db_name + "->", "")\
                .replace(db_name + "-#", "")
            return demultilined_res
        else:
            return ""

    def parse_psql_version(self, psql: str) -> str:
        """Parse for psql version and return version number.

        :param psql: psql --version output

        :returns: version string (e.g "14.5")
        """

        match_version_stmt: ParserElement = \
            Literal("psql (PostgreSQL) ") + Combine(Word(nums) + '.' + Word(nums))
        stmt_res: ParseResults = None
        result: str = ""

        try:
            stmt_res = match_version_stmt.parse_string(psql)
        except ParseException as e:
            if self.debug:
                f = open("psqlparser.log", "a")
                f.write(str(e.explain()) + "\n")
                f.close()

        if stmt_res:
            result = stmt_res.as_list()[1]

        return result



import sys
import re
from dataclasses import dataclass
from typing import Optional

import sqlglot
import sqlglot.expressions as exp
from sqlglot.dialects.postgres import Postgres
import psycopg


@dataclass(frozen=True)
class PostgreSQLDataType:
    name: str
    digits: Optional[int]
    precision: Optional[int]


@dataclass(frozen=True)
class Column:
    name: str
    type: PostgreSQLDataType
    table: str


class SqlParser:
    # Patches the postgres dialect to recognize bpchar
    Postgres.Tokenizer.KEYWORDS["BPCHAR"] = sqlglot.TokenType.CHAR

    def __init__(self, db_connection: psycopg.Connection):
        self.dialect: str = "postgres"
        self.db_connection: psycopg.Connection = db_connection

    def parse(self, sql: str) -> list[sqlglot.exp.Expression]:
        """
        Parses all the statements in 'sql'.

        'sql' should be a string of one or more postgresql statements
        (delimited by ';'). The last ';' in 'sql' is optional.
        Throws sqlglot.ParseError on invalid sql.
        """

        return sqlglot.parse(sql, read=self.dialect)

    def parse_one(self, sql: str) -> sqlglot.exp.Expression:
        """
        Parses the first statement in 'sql'.
        Does not validate that 'sql' contains only 1 statement!
        'sql' should be a postgresql statement.
        The trailing ';' in 'sql' is optional.
        Throws sqlglot.ParseError on invalid sql.
        """

        return sqlglot.parse_one(sql, read=self.dialect)

    def get_query_columns(self, parsed_sql: exp.Expression) -> list[Column]:
        """
        Gets all columns from all tables mentioned in parsed_sql.
        """

        table_names = self.find_all_table_names(parsed_sql)

        columns = []
        for table_name in table_names:
            table_columns = self._get_columns(table_name)
            columns.extend(table_columns)

        return columns

    @staticmethod
    def get_root_node(node: exp.Expression) -> exp.Expression:
        """
        Finds the root node (probably exp.Select) from node.
        """
        root_node = node

        while True:
            new_root = root_node.find_ancestor(exp.Expression)
            if new_root is None:
                break
            root_node = new_root
        return root_node

    @staticmethod
    def find_all_table_names(parsed_sql: exp.Expression) -> list[str]:
        """
        Finds all unique table names in 'parsed_sql'.
        Any table name with an alias is listed without the alias
        (e.g.: FROM <table> AS <alias>)
        Also removes duplicate table names from the result.
        """
        table_names = []
        tables = parsed_sql.find_all(exp.Table)
        # table.this.this: first .this gets the exp.Identifier of the table and
        # second .this gets the name of the identifier.
        table_names = [table.this.this for table in tables]

        unique_table_names = list(set(table_names))

        return unique_table_names

    @staticmethod
    def get_column_name_from_column_expression(column_expression: exp.Column) -> str:
        """
        Returns the column name of column expression ast node.
        """
        return column_expression.find(exp.Identifier).this

    @staticmethod
    def find_where_predicates(root: exp.Expression) -> list[exp.Predicate]:
        """
        Finds all Predicate nodes inside Where statements from some root node
        (usually the whole parsed output from sqlparser.parse_one()).
        This function takes care to not introduce any duplicate Predicate's in
        case of nested Where statements.
        """
        predicates = []

        where_statements = root.find_all(exp.Where)

        toplevel_where_statements = list(
            filter(lambda x: not x.find_ancestor(exp.Where), where_statements)
        )

        for where_statement in toplevel_where_statements:
            for predicate in where_statement.find_all(exp.Predicate):
                predicates.append(predicate)

        return predicates

    def _get_columns(self, table_name: str) -> list[Column]:
        """
        Gets the columns from db 'db_name' table 'table_name'.
        """
        names = self._get_column_names(table_name)
        types = self._get_column_types(table_name, names)

        assert len(names) == len(types)

        columns = []
        i = 0
        for name in names:
            columns.append(Column(name, types[i], table_name))
            i += 1

        return columns

    def _get_column_names(self, table_name: str) -> list[str]:
        """
        Requires that 'table_name' exists.
        Extracts columns of the table by running postgres specific metadata
        query.
        """

        statement = (
            "SELECT column_name "
            "FROM information_schema.columns "
            f"WHERE table_name = '{table_name}';"
        )

        column_names = []
        with self.db_connection.cursor() as cursor:
            cursor.execute(statement)
            column_names = cursor.fetchall()
            self.db_connection.rollback()

        # transforms 1-element tuples to just list of elements
        column_names = [x[0] for x in column_names]
        #        print(f"column_names: {column_names}")

        return column_names

    def _get_column_types(
        self, table_name: str, columns: list[str]
    ) -> list[PostgreSQLDataType]:
        # sql adapted from: https://stackoverflow.com/a/20194807/13540518
        column_type_statement_prefix = f"""
SELECT
    pg_catalog.format_type(a.atttypid, a.atttypmod) AS "data_type"
FROM
    pg_catalog.pg_attribute AS a
WHERE
    a.attnum > 0
    AND NOT a.attisdropped
    AND a.attrelid = (
        SELECT c.oid
            FROM pg_catalog.pg_class c
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE c.relname = '{table_name}'
                AND pg_catalog.pg_table_is_visible(c.oid)
                AND a.attname IN ("""

        if len(columns) == 0:
            return ""

        in_condition = ""
        for col in columns[:-1]:
            in_condition += f"'{col}', "
        in_condition += f"'{columns[-1]}')\n    );"
        column_type_statement = column_type_statement_prefix + in_condition

        type_names = []
        with self.db_connection.cursor() as cursor:
            cursor.execute(column_type_statement)
            type_names = cursor.fetchall()
            self.db_connection.rollback()

        # transforms 1-element tuples to just list of elements
        type_names = [x[0] for x in type_names]
        #        print(f"type_names: {type_names}")

        parseable_type_names = self._convert_from_internal_types(type_names)

        return parseable_type_names

    def _convert_from_internal_types(
        self, type_names: list[str]
    ) -> list[PostgreSQLDataType]:
        """
        Postgresql has internal type names such as: character or character(x).
        sqlglot can only parse the types in the form they are declared (atleast
        in column definitions inside CREATE TABLE) like: CHAR or CHAR(x).
        This function converts from the internal to the declared form.
        """

        converted_types = []
        character_matcher = re.compile(r"^(?:character|char)\(\s*(\d+)\s*\)$")
        varchar_matcher = re.compile(
            r"^(?:(?:character varying)|varchar)\(\s*(\d+)\s*\)$"
        )
        numeric_matcher = re.compile(
            r"^(?:numeric|decimal)\s*\(\s*(\d+)(?:\s*,\s*(\d+))?\s*\)$"
        )

        for type_name in type_names:
            if type_name == "integer":
                converted_types.append(PostgreSQLDataType("INT", None, None))
            elif type_name == "character":
                converted_types.append(PostgreSQLDataType("CHAR", None, None))
            elif type_name == "char":
                converted_types.append(PostgreSQLDataType("CHAR", None, None))
            elif match := character_matcher.match(type_name):
                conv_type = PostgreSQLDataType(
                    f"CHAR({match.group(1)})", int(match.group(1)), None
                )
                converted_types.append(conv_type)

            elif match := varchar_matcher.match(type_name):
                conv_type = PostgreSQLDataType(
                    f"VARCHAR({match.group(1)})", int(match.group(1)), None
                )
                converted_types.append(conv_type)

            elif match := numeric_matcher.match(type_name):
                num_groups = len(match.groups())
                if num_groups == 1:
                    conv_type = PostgreSQLDataType(
                        f"NUMERIC({match.group(1)})", int(match.group(1)), None
                    )
                    converted_types.append(conv_type)

                elif num_groups == 2:
                    name = f"NUMERIC({match.group(1)},{match.group(2)})"
                    conv_type = PostgreSQLDataType(
                        name, int(match.group(1)), int(match.group(2))
                    )
                    converted_types.append(conv_type)
                else:
                    # TODO: proper error handling
                    print(
                        f"unrecognized number '{num_groups}' of arguments for numeric() column type",
                        file=sys.stderr,
                    )
                    exit(1)
            else:
                # TODO: proper error handling
                print(
                    f"unable to convert from internal type '{type_name}' to declared type",
                    file=sys.stderr,
                )
                exit(1)

        return converted_types



def main() -> None:
    import os
    import sys
    from functools import reduce

    from .psqlconninfo import PsqlConnInfo
    from .semanticrouter import SemanticRouter
    from .psqlparser import PsqlParser
    from .psqlwrapper import PsqlWrapper

    if len(sys.argv) > 1:
        conn_info = PsqlConnInfo(
            reduce(lambda x, y: x + y, sys.argv[1:], "")  # concat arguments
            ).get()
        sem_router = SemanticRouter(*conn_info)  # asterisk unpacks the 5-tuple
        psql = PsqlWrapper(sys.argv[1].encode("utf-8"),
                           sem_router.run_analysis,
                           PsqlParser())
        psql.start()
    else:
        print(f"{os.path.basename(sys.executable)} \
                main.py [psql arguments] <database name>")


if __name__ == "__main__":
    main()



import psycopg
from sqlglot import exp
from typing import Optional

from .sqlparser import SqlParser, Column
from .qepparser import QEPAnalysis, QEPParser

# analysis modules
from .cmp_domain_checker import CmpDomainChecker
from .eq_wildcard_checker import EqWildcardChecker
from .implied_expression_checker import ImpliedExpressionChecker
from .strange_having_checker import StrangeHavingChecker
from .subquery_order_by_checker import SubqueryOrderByChecker
from .subquery_select_checker import SubquerySelectChecker
from .sum_distinct_checker import SumDistinctChecker


class SemanticRouter:
    """Analyze given SQL queries via a plethora of analysis modules."""

    def __init__(self,
                 pg_host: str,
                 pg_port: str,
                 pg_user: str,
                 pg_pass: str,
                 pg_name: str
                 ):
        """Initialize Postgres connection with given paramaters."""
        self.pg_host: str = pg_host
        self.pg_port: str = pg_port
        self.pg_user: str = pg_user
        self.pg_pass: str = pg_pass
        self.pg_name: str = pg_name

    def run_analysis(self, sql_query: str) -> str:
        """Run analysis modules on SQL query string and get an insightful \
        message in return.

        Semantic router (some day) implements basic heuristics to avoid
        running all the modules on all queries. For now, it is dumb brute
        force router.
        :param sql_query: is a single well-formed query to run analytics on.
        :returns: an insightful message that might include vt100-compatible \
        control codes. \n is newline (carriage return \r will be added by \
        wrapper).
        """
        with psycopg.connect("host=" + self.pg_host +
                             " port=" + self.pg_port +
                             " dbname=" + self.pg_name +
                             " user=" + self.pg_user +
                             " password=" + self.pg_pass) as conn:
            sql_parser: SqlParser = SqlParser(conn)
            sanitized_sql: exp.Expression = sql_parser.parse_one(sql_query)
            qep_analysis: QEPAnalysis = QEPParser(conn=conn).parse(sql_query)
            analysis_result: Optional[str] = None

            # Comparing different domains
            columns: list[Column] = sql_parser.get_query_columns(sanitized_sql)
            analysis_result = CmpDomainChecker(sanitized_sql, columns).check()

            if analysis_result is not None:
                return analysis_result

            # ORDER BY in subquery
            analysis_result = \
                SubqueryOrderByChecker(sanitized_sql, qep_analysis).check()

            if analysis_result is not None:
                return analysis_result

            # SELECT in subquery
            analysis_result = \
                SubquerySelectChecker(sanitized_sql, sql_parser).check()

            if analysis_result is not None:
                return analysis_result

            # Implied expression
            analysis_result = \
                ImpliedExpressionChecker(sanitized_sql, sql_query,
                                         conn).check()

            if analysis_result is not None:
                return analysis_result

            # Strange HAVING clause without GROUP BY
            analysis_result = \
                StrangeHavingChecker(sanitized_sql, qep_analysis).check()

            if analysis_result is not None:
                return analysis_result

            # SUM/AVG(DISTINCT)
            analysis_result = \
                SumDistinctChecker(sanitized_sql, qep_analysis).check()

            if analysis_result is not None:
                return analysis_result

            # Wildcards without LIKE
            analysis_result = \
                EqWildcardChecker(sanitized_sql, qep_analysis).check()

            if analysis_result is not None:
                return analysis_result

        return ""  # No semantic errors found
